import pandas as pd
from tqdm.notebook import tqdm
import warnings

from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.Time import Time

class CircuitQuery(object):

    def __init__(self, circuit_type, circuit_name):
        self.circuit_type = circuit_type
        self.circuit_name = circuit_name

    # PIC
    def find_timestamp_pic(self, timestamp_fgc, *, spark, duration=[(1, 's'), (60, 's')], verbose=True):
        if verbose:
            print('Querying NXCALS for {} PIC signal of at {} with duration {}.'.
                  format(self.circuit_name, timestamp_fgc, duration))

        st_abort_pic_df = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=int(timestamp_fgc), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PIC', signal='ST_ABORT_PIC') \
            .signal_query().dfs[0]

        return st_abort_pic_df.index[0]

    # PC
    def find_source_timestamp_pc(self, t_start, t_end, verbose=True):
        if verbose:
            print('Querying PM for {} EE timestamp at from {} to {}.'
                  .format(self.circuit_name, t_start, t_end))

        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC') \
            .event_query().df

    def query_pc_pm(self, timestamp_fgc, timestamp_sync, *, signal_names, verbose=True):
        if verbose:
            print('Querying PM for {} FGC signals {} at {}.'
                  .format(self.circuit_name, signal_names, timestamp_fgc))

        pc_pm_dfs = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal=signal_names) \
            .signal_query() \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

        return pc_pm_dfs

    def get_timestamp_ref(self, col='fgcPm'):
        timestamp_ref = Reference.get_power_converter_reference_fpa(self.circuit_type, self.circuit_name, col)
        return Time.to_unix_timestamp(timestamp_ref)

    # EE
    def find_source_timestamp_ee(self, t_start, duration=[(10, 's'), (1000, 's')], verbose=True):
        if verbose:
            print('Querying PM for {} EE timestamp at {} with duration {}.'
                  .format(self.circuit_name, t_start, duration))

        return QueryBuilder().with_pm() \
            .with_duration(t_start=int(t_start), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='EE') \
            .event_query().df

    def query_ee_u_dump_res_pm(self, timestamp_ee, timestamp_sync, system, *, signal_names, verbose=True):
        if verbose:
            print('Querying PM for {} EE voltage signals {} at {}.'
                  .format(self.circuit_name, signal_names, timestamp_ee))

        ee_pm_dfs = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='EE', signal=signal_names) \
            .signal_query() \
            .remove_initial_offset(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

        return ee_pm_dfs

    def query_ee_t_res_pm(self, timestamp_ee, timestamp_sync, system, *, signals, verbose=True):
        if verbose:
            print('Querying PM for {} EE temperature signals {} at {}.'
                  .format(self.circuit_name, signals, timestamp_ee))

        ee_t_res_pm_dfs = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='EE', signal=signals) \
            .signal_query() \
            .remove_values_for_time_less_than(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

        return ee_t_res_pm_dfs

    def query_ee_cals(self, timestamp_fgc, system, *, signal_names, ldb, duration=[(3 * 3600, 's')],
                      t_thr=1000, verbose=True):
        if verbose:
            print('Querying CALS for {} EE temperature signals {} at {} with duration {}.'
                  .format(self.circuit_name, signal_names, timestamp_fgc, duration))

        ee_cals_dfs = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='EE', signal=signal_names) \
            .signal_query() \
            .synchronize_time(timestamp_fgc / 1e9) \
            .remove_values_for_time_less_than(t_thr) \
            .dfs

        return ee_cals_dfs

    # iQPS
    def find_source_timestamp_qds(self, t_start, duration=[(10, 's'), (400, 's')], verbose=True):
        if verbose:
            print('Querying PM for {} iQPS timestamps at {} with duration {}.'
                  .format(self.circuit_name, t_start, duration))

        return QueryBuilder().with_pm() \
            .with_duration(t_start=int(t_start), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query() \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    def query_iqps_analog_pm(self, source_timestamp_qds_df, signals, verbose=True):
        if verbose:
            print('Querying PM for {} iQPS analog signals {}.'.format(self.circuit_name, signals))

        iqps_analog_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_analog_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signals, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec() \
                .filter_median().dfs
            iqps_analog_dfs.append(iqps_analog_df)

        return iqps_analog_dfs

    def query_iqps_digital_pm(self, source_timestamp_qds_df, signals, verbose=True):
        if verbose:
            print('Querying PM for {} iQPS analog signals {}.'
                  .format(self.circuit_name, signals))

        iqps_digital_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_digital_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signals, source=source_qds) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec().dfs
            iqps_digital_dfs.append(iqps_digital_df)

        return iqps_digital_dfs

    # nQPS
    def find_source_timestamp_nqps(self, t_start, duration=[(10, 's'), (400, 's')], verbose=True):
        if verbose:
            print('Querying PM for {} nQPS timestamps at {} with duration {}.'
                  .format(self.circuit_name, t_start, duration))

        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
        return QueryBuilder().with_pm() \
            .with_duration(t_start=int(t_start), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source='*') \
            .event_query() \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    def query_nqps_voltage_pm(self, source_timestamp_qds_df, verbose=True):
        if verbose:
            print('Querying PM for {} nQPS analog signals.'.format(self.circuit_name))

        u_nqps_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']
            qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
            system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
            signal = 'U_DIODE_{}'.format(self.circuit_name.split('.')[0])
            wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

            source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)
            if not source_timestamp_nqps_df.empty:
                timestamp_nqps = source_timestamp_nqps_df.loc[0, 'timestamp']

                u_diode_rqd_df = QueryBuilder().with_pm() \
                    .with_timestamp(timestamp_nqps) \
                    .with_circuit_type(self.circuit_type) \
                    .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, source=qps_crate,
                                                  wildcard={'MAGNET': wildcard}) \
                    .signal_query() \
                    .synchronize_time(timestamp_nqps) \
                    .convert_index_to_sec().dfs[0]

                u_ref_rqd_df = QueryBuilder().with_pm() \
                    .with_timestamp(timestamp_nqps) \
                    .with_circuit_type(self.circuit_type) \
                    .with_metadata(circuit_name=self.circuit_name, system=system, signal='U_REF_N1', source=qps_crate,
                                   wildcard={'QPS_CRATE': qps_crate}) \
                    .signal_query() \
                    .synchronize_time(timestamp_nqps) \
                    .convert_index_to_sec().dfs[0]
                u_nqps_dfs.append([u_diode_rqd_df, u_ref_rqd_df])
            else:
                warnings.warn('nQPS buffer is empty for %s !' % source_qds)
                u_nqps_dfs.append([pd.DataFrame(), pd.DataFrame()])

        return u_nqps_dfs

    # QH
    def query_qh_pm(self, source_timestamp_qds_df, *, signal_names, is_ref=False, verbose=True):
        if verbose:
            print('Querying PM for {} QH analog signals {}.'.format(self.circuit_name, signal_names))

        hds_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            source_qds = row['source']
            if is_ref:
                timestamp_qh = Reference.get_quench_heater_reference_discharge(self.circuit_type, source_qds)
            else:
                timestamp_qh = row['timestamp'] + 1000000

            hds_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qh) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QH', signal=signal_names, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qh) \
                .convert_index_to_sec().dfs

            hds_dfs.append(hds_df)

        return hds_dfs

    # DIODE LEADS
    def find_source_timestamp_nqps_pm(self, source_qds, timestamp_qds, duration=[(600, 's'), (600, 's')], verbose=True):
        if verbose:
            print('Querying PM for {} nQPS timestamps at {} with duration {}.'
                  .format(self.circuit_name, timestamp_qds, duration))

        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])

        return QueryBuilder().with_pm() \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source=qps_crate) \
            .event_query().df

    def _query_i_u_diode_pm(self, circuit_name, source_qds, timestamp_fgc, timestamp_qds_a, timestamp_qds_b,
                            verbose=True):
        if verbose:
            print('Querying PM for {} DIODE analog signals.'.format(self.circuit_name))

        i_a_rqd_df = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_A') \
            .signal_query().dfs[0]

        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_a_rqd_df = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_diode_b_rqd_df = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_b) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_ref_rqd_df = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal='U_REF_N1', source=qps_crate,
                           wildcard={'QPS_CRATE': qps_crate}) \
            .signal_query().dfs[0]

        return i_a_rqd_df, u_diode_a_rqd_df, u_diode_b_rqd_df, u_ref_rqd_df

    def _query_i_u_diode_nxcals(self, circuit_name, source_qds, timestamp_qds, *, spark,
                                duration=[(50, 's'), (150, 's')], verbose=True):
        if verbose:
            print('Querying NXCALS for {} DIODE analog signals at {} with duration {}.'
                  .format(self.circuit_name, timestamp_qds, duration))

        i_meas_diode_df = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=int(timestamp_qds), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_MEAS') \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_rqd_df = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=int(timestamp_qds), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        return i_meas_diode_df, u_diode_rqd_df

    def query_current_voltage_diode_leads_pm(self, timestamp_fgc_rqd, results_table, verbose=True):

        i_meas_u_diode_u_ref_pm_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0], desc='Querying PM'):
            source_qds, timestamp_qds, u_diode_db = row[['magnet', 'timestamp_iqps', 'nqps_db']]

            if u_diode_db == 'PM':
                source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)
                timestamp_nqps_board_a = source_timestamp_nqps_df.loc[0, 'timestamp']
                timestamp_nqps_board_b = source_timestamp_nqps_df.loc[1, 'timestamp']
                i_meas_u_diode_u_ref_pm_df = self._query_i_u_diode_pm(self.circuit_name, source_qds, timestamp_fgc_rqd,
                                                                      timestamp_nqps_board_a, timestamp_nqps_board_b,
                                                                      verbose)
            else:
                i_meas_u_diode_u_ref_pm_df = pd.DataFrame()

            i_meas_u_diode_u_ref_pm_dfs.append(i_meas_u_diode_u_ref_pm_df)

        return i_meas_u_diode_u_ref_pm_dfs

    def query_current_voltage_diode_leads_nxcals(self, results_table, *, spark, verbose=True):

        i_meas_u_diode_nxcals_dfs = []
        for index, row in results_table.iterrows():
            source_qds, timestamp_qds, u_diode_db = row[['magnet', 'timestamp_iqps', 'nqps_db']]

            #ToDo fix t_start=int(timestamp_qds)
            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            i_meas_u_diode_nxcals_df = self._query_i_u_diode_nxcals(self.circuit_name, source_qds,
                                                                    timestamp_qds, spark=spark, verbose=verbose)

            i_meas_u_diode_nxcals_dfs.append(i_meas_u_diode_nxcals_df)

        return i_meas_u_diode_nxcals_dfs

    # U_DIODE, U_EARTH
    def query_voltage_cals(self, system, signal, timestamp_fgc, *, ldb, duration=[(50, 's'), (150, 's')],
                           verbose=True):

        if verbose:
            print('Querying CALS for {} {} signals {} at {} with duration {}.'
                  .format(self.circuit_name, system, signal, timestamp_fgc, duration))

        u_cals_dfs = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=int(timestamp_fgc), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, wildcard={'MAGNET': '*'}) \
            .signal_query() \
            .synchronize_time(timestamp_fgc / 1e9).dfs

        return u_cals_dfs

    # LEADS
    def find_timestamp_leads(self, timestamp_fgc_rq, system='LEADS', duration=[(10, 's'), (400, 's')], verbose=True):
        if verbose:
            print('Querying PM for {} DFB timestamps at {} with duration {}.'.
                  format(self.circuit_name, timestamp_fgc_rq, duration))

        return QueryBuilder().with_pm() \
            .with_duration(t_start=int(timestamp_fgc_rq), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system) \
            .event_query().df

    def query_leads(self, timestamp_fgc, source_timestamp_leads_df, *,
                    system='LEADS', signals, spark, duration=[(30, 's'), (90, 's')], verbose=True):

        if source_timestamp_leads_df.empty:
            return self._query_leads_nxcals(duration, signals, spark, system, timestamp_fgc, verbose)
        else:
            return self._query_leads_pm(signals, source_timestamp_leads_df, system, timestamp_fgc, verbose)

    def _query_leads_nxcals(self, duration, signal_names, spark, system, timestamp_fgc, verbose=True):
        if verbose:
            print('Querying NXCALS for {} DFB signals {} at {} with duration {}.'
                  .format(self.circuit_name, signal_names, timestamp_fgc, duration))

        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=int(timestamp_fgc), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query() \
            .synchronize_time(timestamp_fgc) \
            .convert_index_to_sec().dfs

    def _query_leads_pm(self, signal_names, source_timestamp_leads_df, system, timestamp_fgc, verbose):
        if verbose:
            print('Querying PM for {} DFB signals {} at {}.'.format(self.circuit_name, signal_names, timestamp_fgc))

        source_leads = source_timestamp_leads_df.loc[0, 'source']
        timestamp_leads = source_timestamp_leads_df.loc[0, 'timestamp']
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_leads) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system,
                           signal=signal_names, source=source_leads) \
            .signal_query() \
            .synchronize_time(timestamp_leads) \
            .convert_index_to_sec().dfs