import matplotlib.pyplot as plt
from IPython.display import display, HTML
import pandas as pd

from lhcsmapi.Time import Time
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis, convert_plot_features_map_to_text
from lhcsmapi.analysis.magnet.iqds.IqdsRqAnalysis import find_circuit_name_with_first_quenched_magnet, choose_u_qs0, \
    find_t_start_t_end_voltage_slope, calculate_voltage_slope
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.reference.Reference import Reference


def plot_voltage_qh(source_qds, timestamp_qds, timestamp_ref, u_hds_rq_df, u_hds_rq_ref_df, is_log=False):
    fig, ax = plt.subplots(figsize=(15, 7.5))
    title = "Magnet: {}, Time Stamp: {}, Reference {}, U_HDS(t)"\
        .format(source_qds, Time.to_string_short(timestamp_qds + 1000000), Time.to_string_short(timestamp_ref))
    fig.suptitle(title, fontsize=20)
    u_hds_rq_df[0].plot(ax=ax, grid=True)
    u_hds_rq_df[1].plot(ax=ax, grid=True)
    u_hds_rq_ref_df[0].plot(ax=ax, grid=True, style='--')
    u_hds_rq_ref_df[1].plot(ax=ax, grid=True, style='--')
    ax.tick_params(labelsize=15)
    ax.set_ylabel("U_HDS, [V]", color='C0', fontsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)
    if is_log:
        plt.yscale('log')
    plt.show()


def plot_single_ee_fgc(u_dump_res_df, i_meas_df, *, title='', plot_text=''):

    ax1 = i_meas_df.plot(title=title, color='C0', legend=True, grid=True)
    ax1.title.set_size(20)
    ax1.set_ylabel("I_MEAS, [A]", color='C0', fontsize=15)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.legend(loc='lower left')
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    u_dump_res_df.plot(ax=ax2, figsize=(15, 7.5), color='C2', legend=True)
    ax2.set_ylabel("U_DUMP_RES, [V]", color='C2', fontsize=15)

    ax2.annotate(plot_text, xytext=(max(u_dump_res_df.index.max() * 1.2, i_meas_df.index.max()) * 1.1, 0),
                 xy=(0.6, 0), fontsize=15, family='monospace')
    ax2.legend(loc='upper right')
    ax2.tick_params(labelsize=15)
    ax2.set_xlim((-50, ax2.get_xlim()[1]))
    plt.show()


class RqCircuitAnalysis(CircuitAnalysis):

    def analyze_qds(self, circuit_names, iqps_analog_dfs, iqps_digital_dfs, u_nqps_rqd_dfs, u_nqps_rqf_dfs):
        for index, row in self.results_table.iterrows():
            source_qds, timestamp_qds = row[['magnet', 'timestamp_iqps']]
            magnet_name = 'MQ.{}'.format(source_qds)

            u_qs0_ext_rq_df, u_qs0_int_rq_df, u_1_ext_rq_df, u_2_ext_rq_df, u_1_int_rq_df, u_2_int_rq_df = \
                iqps_analog_dfs[
                    index]
            st_magnet_ok_ext_rq_df, st_magnet_ok_int_rq_df, st_nqd0_ext_rq_df, st_nqd0_int_rq_df = iqps_digital_dfs[
                index]

            u_diode_rqd_df, u_ref_rqd_df = u_nqps_rqd_dfs[index]
            u_diode_rqf_df, u_ref_rqf_df = u_nqps_rqf_dfs[index]

            # Analysis
            circuit_name = find_circuit_name_with_first_quenched_magnet(circuit_names, u_diode_rqd_df, u_diode_rqf_df)
            aperture = MappingMetadata.get_rq_aperture(circuit_name, magnet_name) if len(circuit_name) > 0 else ''
            u_qs0_aperture_df = choose_u_qs0(aperture, u_qs0_int_rq_df, u_qs0_ext_rq_df) if len(aperture) > 0 \
                                                                                        else pd.DataFrame()
            t_slope_start, t_slope_end = find_t_start_t_end_voltage_slope(u_qs0_aperture_df)
            du_dt = calculate_voltage_slope(u_qs0_aperture_df, t_slope_start, t_slope_end)
            if len(aperture) > 0:
                print('The first aperture to quench is {} of magnet {}'.format(aperture, magnet_name))
                self.results_table.loc[index, 'origin'] = '{}/{}'.format(circuit_name.split('.')[0], aperture)
                print('The voltage increase is {:.3f} V/s'.format(du_dt))
            else:
                print('No aperture quenched in magnet {}'.format(magnet_name))
                self.results_table.loc[index, 'origin'] = 'None'

            self.results_table.loc[index, 'dU_QS0/dt'] = round(du_dt, 2)

            fig, ax = plt.subplots(2, 2, figsize=(30, 15))
            title = "Magnet: {}, Aperture: {}/{}, Time Stamp: {}".format(source_qds, circuit_name.split('.')[0],
                                                                         aperture, Time.to_string_short(timestamp_qds))
            fig.suptitle(title, fontsize=30)

            # iQPS analog
            if (not u_qs0_ext_rq_df.empty) and (not u_qs0_int_rq_df.empty):
                u_qs0_ext_rq_df.plot(ax=ax[0][0])
                u_qs0_int_rq_df.plot(ax=ax[0][0], grid=True)
                ax[0][0].tick_params(labelsize=15)
                ax[0][0].set_xlim((-0.1, 0.1))
                ax[0][0].set_ylim((-0.3, 0.3))
                ax[0][0].axhspan(ymin=-0.1, ymax=+0.1, facecolor='xkcd:yellowgreen')
                ax[0][0].axvspan(xmin=t_slope_start, xmax=t_slope_end, facecolor='xkcd:goldenrod')

            # nQPS analog
            try:
                u_diode_rqd_df.plot(ax=ax[0][1])
                u_ref_rqd_df.plot(ax=ax[0][1])
                u_diode_rqf_df.plot(ax=ax[0][1])
                u_ref_rqf_df.plot(ax=ax[0][1], grid=True)
            except Exception as e:
                pass
            ax[0][1].tick_params(labelsize=15)

            # iQPS digital
            try: # ToDo: Replace with a check if signal is not empty
                st_magnet_ok_ext_rq_df.astype(int).plot(ax=ax[1][0])
                st_magnet_ok_int_rq_df.astype(int).plot(ax=ax[1][0], grid=True)
            except Exception as e:
                pass
            ax[1][0].tick_params(labelsize=15)
            ax[1][0].set_xlim((-0.1, 0.1))
            st_nqd0_ext_rq_df.astype(int).plot(ax=ax[1][0])
            st_nqd0_int_rq_df.astype(int).plot(ax=ax[1][0], grid=True)
            ax[1][0].tick_params(labelsize=15)

            # nQPS analog - zoom
            try:
                u_diode_rqd_df.plot(ax=ax[1][1])
                u_ref_rqd_df.plot(ax=ax[1][1])
                u_diode_rqf_df.plot(ax=ax[1][1])
                u_ref_rqf_df.plot(ax=ax[1][1], grid=True)
            except Exception as e:
                pass
            ax[1][1].tick_params(labelsize=15)
            ax[1][1].set_xlim((-0.1, 0.1))
            plt.show()

    def analyze_qh(self, circuit_type, u_hds_rq_dfs, u_hds_rq_ref_dfs):
        for index, row in self.results_table.iterrows():
            source_qds, timestamp_qds = row[['magnet', 'timestamp_iqps']]
            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            timestamp_ref = Reference.get_quench_heater_reference_discharge(circuit_type, source_qds)

            u_hds_rq_df = u_hds_rq_dfs[index]
            u_hds_rq_ref_df = u_hds_rq_ref_dfs[index]

            # Calculate features
            # # to reference values
            first_last_u_df = FeatureBuilder().with_signal(u_hds_rq_df) \
                .calculate_features(features=['first', 'last20mean'], index=timestamp_qds)

            is_first_last_u_correct = Reference.compare_features_to_reference(first_last_u_df, self.circuit_type,
                                                                                  'QH',
                                                                              wildcard={'CELL': source_qds})

            # # to a reference discharge
            tau_df = FeatureBuilder().with_signal(u_hds_rq_df) \
                .calculate_features(features=['tau_charge'], index=timestamp_qds)

            tau_ref_df = FeatureBuilder().with_signal(u_hds_rq_ref_df) \
                .calculate_features(features=['tau_charge'], index=timestamp_ref)

            is_tau_correct = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df,
                                                                                   self.circuit_type, 'QH',
                                                                                   wildcard={'CELL': source_qds})

            # Update results table
            # self.results_table.loc[index, 'QH'] = is_first_last_u_correct and is_tau_correct
            # Use OK rather than True always
            self.results_table.loc[index, 'QH'] = 'OK' if is_first_last_u_correct and is_tau_correct else 'NOT_OK'

            # Display features
            features_qh_concat = pd.concat([tau_df, tau_ref_df], axis=0)
            display(HTML(features_qh_concat.to_html()))
            display(HTML(first_last_u_df.to_html()))

            # Plot
            plot_voltage_qh(source_qds, timestamp_qds, timestamp_ref, u_hds_rq_df, u_hds_rq_ref_df)
            plot_voltage_qh(source_qds, timestamp_qds, timestamp_ref, u_hds_rq_df, u_hds_rq_ref_df, is_log=True)

    def create_mp3_results_table(self, circuit_names, timestamp_fgc_rqd):

        def create_pm_file_name(timestamp, name):
            return Time.to_string_short(timestamp).replace(' ', '_').replace(':', '').replace('-', '') + '_' + name

        mp3_table_df = Reference.read_mp3_quench_database_columns(self.circuit_type)

        # Circuit
        mp3_table_df['Circuit'] = self.results_table['circuit']
        # Date
        mp3_table_df['Date'] = Time.to_string_short(timestamp_fgc_rqd).split(' ')[0]
        # Time
        mp3_table_df['Time'] = Time.to_string_short(timestamp_fgc_rqd).split(' ')[1].split('.')[0]
        # FGC PM file
        pc_name = SignalMetadata.get_circuit_name_metadata(self.circuit_type, circuit_names[0])['PC']['%PC%']
        mp3_table_df['FGC PM file'] = create_pm_file_name(timestamp_fgc_rqd, pc_name).replace('RQD', 'RQx')
        # Position
        mp3_table_df['Position'] = self.results_table['magnet']
        # iQPS PM file
        mp3_table_df['iQPS PM file'] = self.results_table['timestamp_iqps']
        mp3_table_df['iQPS PM file'] = mp3_table_df.apply(
            lambda row: create_pm_file_name(row['iQPS PM file'], row['Position']), axis=1)
        # I_q_D
        mp3_table_df['I_q_D'] = self.results_table['I_MEAS_RQD_quench']
        # I_q_F
        mp3_table_df['I_q_F'] = self.results_table['I_MEAS_RQF_quench']
        # Origin
        mp3_table_df['Origin'] = self.results_table['origin']
        # R_lead_max_RQD
        mp3_table_df['R_lead_max_RQD'] = self.results_table['R_DIODE_LEADS_RQD']
        # R_lead_max_RQF
        mp3_table_df['R_lead_max_RQF'] = self.results_table['R_DIODE_LEADS_RQF']

        return mp3_table_df
