import warnings
import numpy as np
import pandas as pd

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


class RqCircuitQuery(CircuitQuery):

    def create_report_analysis_template(self, source_timestamp_qds_df, source_timestamp_nqps_df, timestamp_fgc_rqd,
                                        i_meas_rqd_df, i_meas_rqf_df, author):
        COLUMN_NAMES = ['circuit', 'timestamp_pic_rqd', 'timestamp_pic_rqf', 'timestamp_fgc_rqd', 'timestamp_fgc_rqf',
                        'timestamp_ee_rqd', 'timestamp_ee_rqf', 'timestamp_leads_rqd', 'timestamp_leads_rqf',
                        'datetime_fgc', 'timestamp_iqps', 'timestamp_nqps', 'iqps_or_nqps_trigger', 'I_MEAS_RQD_quench',
                        'I_MEAS_RQF_quench', 'magnet', 'nqps_crate', 'origin', 'dU_QS0/dt', 'QH', 'R_DIODE_LEADS_RQD',
                        'R_DIODE_LEADS_RQF', 'nqps_db', 'comments', 'performed_by', 'lhcsmapi_version']

        report_qds = pd.DataFrame(columns=COLUMN_NAMES)

        for index, row in source_timestamp_qds_df.iterrows():
            timestamp_qds = row['timestamp']
            qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, row['source'])
            if source_timestamp_nqps_df.empty:
                timestamp_nqps = np.nan
            else:
                mask = source_timestamp_nqps_df['source'] == qps_crate
                timestamp_nqps = source_timestamp_nqps_df[mask]['timestamp'].values[0]
            iqps_or_nqps_trigger = 'nQPS' if timestamp_nqps < timestamp_qds else 'iQPS'

            t_quench = (min(timestamp_qds, timestamp_nqps) - timestamp_fgc_rqd) * 1e-9
            if t_quench > i_meas_rqd_df.index[-1]:
                i_quench_rqd = i_meas_rqd_df.values[-1]
                i_quench_rqf = i_meas_rqf_df.values[-1]
                warnings.warn('Quench time beyond PM buffer. Need to query NXCALS in order to get the quench current!')
            else:
                i_quench_rqd = int(i_meas_rqd_df.reindex([t_quench], method='nearest', tolerance=.1).values[0])
                i_quench_rqf = int(i_meas_rqf_df.reindex([t_quench], method='nearest', tolerance=.1).values[0])

            nqps_db = 'PM' if t_quench < 2 else 'NXCALS'
            values = [self.circuit_name.replace('D', ''), np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan,
                      np.nan, Time.to_string_short(timestamp_fgc_rqd), Time.to_string_short(timestamp_qds),
                      timestamp_nqps, iqps_or_nqps_trigger, i_quench_rqd, i_quench_rqf, row['source'], qps_crate, '',
                      np.nan, '', np.nan, np.nan, nqps_db, '', author, lhcsmapi.__version__]
            report_qds.loc[len(report_qds.index)] = values

        return report_qds

    def query_voltage_cals(self, system, signal, timestamp_fgc, *, ldb, duration=[(50, 's'), (150, 's')],
                           verbose=True):
        u_cals_dfs = super().query_voltage_cals(system, signal, timestamp_fgc, ldb=ldb, duration=duration,
                                                verbose=verbose)

        if 'DIODE' in system:
            rq_diode = MappingMetadata.read_magnet_to_cell_qps_crate(circuit_type='RQ')
            crates = list(rq_diode[rq_diode['Crate U_DIODE_RQx'].isna() & (rq_diode['Circuit'] == self.circuit_name)][
                              'Crate U_REF_N1'].values)

            u_ref_dfs = QueryBuilder().with_cals(ldb) \
                .with_duration(t_start=int(timestamp_fgc), duration=duration) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system=system, signal='U_REF_N1',
                               wildcard={'QPS_CRATE': crates}) \
                .signal_query() \
                .synchronize_time(timestamp_fgc / 1e9).dfs

            return u_cals_dfs + u_ref_dfs

        return u_cals_dfs
