import matplotlib.pyplot as plt
from copy import deepcopy
import numpy as np
import pandas as pd

from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis, get_expert_comment
from lhcsmapi.dbsignal.SignalProcessing import SignalProcessing


def find_time_of_i_a_i_ref_deviation(i_ref_df, i_a_df):
    # re-interpolate both signals
    idx = i_ref_df.index.union(i_a_df.index)
    signal_comparison_df = i_ref_df.reindex(idx).interpolate('index')
    signal_comparison_df = signal_comparison_df.join(i_a_df).dropna()
    signal_comparison_df[(signal_comparison_df.index > -1) & (signal_comparison_df.index < i_ref_df.idxmax().values[0])]

    # find maximum difference between the time of maximum I_REF value and 1 second before
    t_i_ref_max = i_ref_df.idxmax().values[0]
    diff_df = abs(signal_comparison_df['STATUS.I_REF'] - signal_comparison_df['IAB.I_A'])
    max_diff = diff_df[(diff_df.index > (t_i_ref_max - 1)) & (diff_df.index < t_i_ref_max)].max()

    # find time of start of deviation from maximum difference between maximum value of I_REF and 0 (FGC timestamp)
    mask = (diff_df.index > t_i_ref_max) & (diff_df.index < 0) & (diff_df.values > max_diff)

    return diff_df[mask].idxmin()


def flip_signal(u_qs0_ref):
    # # if there is only a negative peak, then flip the signal
    # # if there are two peaks and the first is negative, then flip the signal
    # # otherwise, do nothing
    # # added the case when U_QS0 is saturated only to negative!
    if (all(u_qs0_ref.max() < 0.2) and all(u_qs0_ref.min() < -0.2)):
        return -deepcopy(u_qs0_ref)
    elif (all(u_qs0_ref.max() > 0.2) & all(u_qs0_ref.min() < -0.2) & (u_qs0_ref.idxmin() < u_qs0_ref.idxmax())):
        return -deepcopy(u_qs0_ref)
    else:
        return deepcopy(u_qs0_ref)


class _600ACircuitAnalysis(CircuitAnalysis):

    # PC
    def analyze_i_meas_pc(self, circuit_name, timestamp_fgc, i_meas_df, i_ref_df, i_a_df):
        title = "%s, %s: I_MEAS(t), I_REF(t), I_A(t)" % (Time.to_string_short(timestamp_fgc), circuit_name)

        ax = i_meas_df.plot(figsize=(20, 7), title=title)
        i_ref_df.plot(ax=ax)
        i_a_df.plot(ax=ax)

        ax.grid(True)
        ax.title.set_size(20)
        ax.tick_params(labelsize=15)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("I, [A]", fontsize=15)

        ax = i_meas_df.plot(figsize=(20, 7), title=title)
        i_ref_df.plot(ax=ax, marker='.')
        i_a_df.plot(ax=ax)

        ax.grid(True)
        ax.title.set_size(20)
        ax.tick_params(labelsize=15)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("I, [A]", fontsize=15)
        ax.set_xlim(-0.25, 1)
        plt.show()

        i_meas_df_der = i_meas_df[i_meas_df.index < i_meas_df.idxmax().values[0]]
        SignalProcessing.calculate_time_derivative_of_signals(i_meas_df_der, i_meas_df_der.index, ['STATUS.I_MEAS'],
                                                              ['dSTATUS.I_MEAS'])

        didt_max = i_meas_df_der['dSTATUS.I_MEAS'].max()
        if i_meas_df_der['dSTATUS.I_MEAS'].max() > 0.1:
            self.results_table['Ramp Rate'] = didt_max
            self.results_table['Plateau Duration'] = 0
        else:
            self.results_table['Ramp Rate'] = 0
            self.results_table['Plateau Duration'] = didt_max

        self.results_table['Quench Current'] = i_meas_df.iloc[i_meas_df.index.get_loc(0, 'nearest')].values[0]

        # ToDo: Ramp duration - if there is a ramp of more than 0.1 A/s, then the plateau is NaN;
        # otherwise calculate plateau and set the di/dt to NaN

        # I_MEAS MIITs
        t_deviation = find_time_of_i_a_i_ref_deviation(i_ref_df, i_a_df)
        i_meas_pos_df = i_meas_df[i_meas_df.index > t_deviation]
        miits_i_meas = np.trapz(i_meas_pos_df.values.reshape(len(i_meas_pos_df.index)) ** 2, i_meas_pos_df.index) * 1e-6
        self.results_table['I_MEAS MIITs'] = miits_i_meas

        print('The I_MEAS(t) MIITs is %s MA^2s.' % miits_i_meas)

    def analyze_i_earth_pc(self, circuit_name, timestamp_fgc, i_earth_df):
        title = "%s, %s: I_EARTH(t)" % (Time.to_string_short(timestamp_fgc), circuit_name)

        ax = i_earth_df.plot(figsize=(15, 7.5), grid=True, title=title)
        ax.legend(loc='lower left')
        ax.title.set_size(20)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("IEARTH, [A]", fontsize=15, color='C0')
        ax.tick_params(labelsize=15)
        plt.show()

        self.results_table['Earth Current'] = i_earth_df.max()

        print('The maximum earth current is %s A.' % i_earth_df.max().values[0])

    # EE
    def analyze_u_dump_res_ee(self, circuit_name, timestamp_fgc, i_meas_df, u_dump_res_df):
        title = "%s, %s: I_EARTH(t)" % (Time.to_string_short(timestamp_fgc), circuit_name)

        ax = i_meas_df.plot(figsize=(20, 7), grid=True, color='C0', title=title)
        ax.legend(loc='lower left')
        ax.title.set_size(20)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("I_MEAS, [A]", fontsize=15, color='C0')
        ax.tick_params(labelsize=15)

        ax1 = ax.twinx()
        u_dump_res_df.plot(ax=ax1, color='C2')
        ax1.legend(loc='upper right')
        ax1.set_ylabel("U_DUMP_RES, [V]", fontsize=15, color='C2')
        ax1.tick_params(labelsize=15)
        ax1.set_xlim(-1, 2)
        plt.show()

        self.results_table['EE max voltage'] = u_dump_res_df.max()

        print('The maximum EE voltage is %s V.' % u_dump_res_df.max().values[0])

    # QPS
    def analyze_qds(self, circuit_name, timestamp_qps, i_dcct_df, i_didt_df, u_diff_df, u_res_df):
        title = "%s, %s: U_RES(t), U_DIFF(t), I_DCCT(t), I_DIDT(t)" % (
            Time.to_string_short(timestamp_qps), circuit_name)

        ax = i_dcct_df.plot(figsize=(20, 7), color='C0', title=title)
        i_didt_df.plot(ax=ax, grid=True, color='C1')
        ax.legend(loc='lower left')
        ax.title.set_size(20)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("I, [A]", fontsize=15, color='C0')
        ax.tick_params(labelsize=15)

        ax1 = ax.twinx()
        u_diff_df.plot(ax=ax1, color='C2')
        u_res_df.plot(ax=ax1, color='C3')
        ax1.legend(loc='upper right')
        ax1.tick_params(labelsize=15)
        ax1.set_ylabel("U, [V]", fontsize=15, color='C2')

        # Get expert input on reason
        self.results_table['Reason'] = get_expert_comment('Reason for quench', self.is_automatic)

        # Get expert input on quench origin
        self.results_table['Quench Origin'] = get_expert_comment('Origin of a quench', self.is_automatic)

    def calculate_u_res_slope(self, circuit_name, timestamp_qps, u_res_df):
        title = "%s, %s: U_RES(t), U_DIFF(t), I_DCCT(t), I_DIDT(t)" % (
        Time.to_string_short(timestamp_qps), circuit_name)

        # Skip first ten points to avoid initial noise
        u_res_slice_df = u_res_df[u_res_df.index > u_res_df.index[10]]
        # Take absolute value
        u_res_flip_df = flip_signal(u_res_slice_df)

        # Find first index of argmax (0.25 V)
        t_u_max = u_res_flip_df[u_res_flip_df >= 0.25].dropna().index[0]
        v_u_max = u_res_df[u_res_df.index == t_u_max].dropna().values[0]

        # Find index of voltage equal to 0.05 V
        t_u_min = u_res_flip_df[(u_res_flip_df.index > t_u_max) & (u_res_flip_df.values < 0.05)].index[-1]
        v_u_min = u_res_df[u_res_df.index == t_u_min].dropna().values[0][0]

        # Calculate dU_RES/dt
        d_u_res_dt = (v_u_max - v_u_min) / (t_u_max - t_u_min)

        self.results_table['dUres/dt'] = d_u_res_dt[0]

        ax = u_res_df[u_res_df.index > u_res_df.index[10]].plot(figsize=(15, 7.5), grid=True, title=title)
        ax.legend(loc='lower left')
        ax.title.set_size(20)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("U_RES, [V]", fontsize=15, color='C0')
        ax.tick_params(labelsize=15)
        slope_df = pd.DataFrame(data=[v_u_min, v_u_max[0]], index=[t_u_min, t_u_max], columns=['dU_RES/dt'])
        slope_df.plot(ax=ax, grid=True)
        ax.legend([u_res_df.columns[0], slope_df.columns[0]])
        plt.show()

        print('The slope of the U_RES signal is %s V/s.' % d_u_res_dt[0])

    def create_mp3_results_table(self):
        columns = ['Circuit Name', 'Family', 'Period', 'Date', 'Time', 'flagged in APEX', 'Quench Current',
                   'DCCT MIITs', 'Reason', 'Pivot', 'Quench Origin',
                   'Ramp Rate', 'Plateau Duration', 'dUres/dt', 'EE max voltage', 'Earth Current', 'Coupling',
                   'Comment', 'Person who analyzed', 'Reference Doc']
        mp3_table = pd.DataFrame(columns=columns, index=[0])

        mp3_table['Circuit Name'] = self.results_table['Circuit Name']
        mp3_table['Family'] = self.results_table['Family']
        mp3_table['Period'] = self.results_table['Period']
        mp3_table['Date'] = self.results_table['Time']
        mp3_table['flagged in APEX'] = self.results_table['flagged in APEX']
        mp3_table['Quench Current'] = self.results_table['Quench Current']
        mp3_table['DCCT MIITs'] = self.results_table['I_MEAS MIITs']
        mp3_table['Reason'] = self.results_table['Reason']
        mp3_table['Pivot'] = self.results_table['Pivot']
        mp3_table['Quench Origin'] = self.results_table['Quench Origin']
        mp3_table['Ramp Rate'] = self.results_table['Ramp Rate']
        mp3_table['Plateau Duration'] = self.results_table['Plateau Duration']
        mp3_table['dUres/dt'] = self.results_table['dUres/dt']
        mp3_table['EE max voltage'] = self.results_table['EE max voltage']
        mp3_table['Coupling'] = self.results_table['Coupling']
        mp3_table['Comment'] = self.results_table['Comment']
        mp3_table['Person who analyzed'] = self.results_table['performed_by']
        mp3_table['Reference Doc'] = self.results_table['Reference Doc']

        return mp3_table

