import pandas as pd
import numpy as np

from pyspark.sql.types import IntegerType
import pyspark.sql.functions as F

from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.pyedsl.ResistanceBuilder import ResistanceBuilder


def generate_busbar_status(beam_mode_with_pattern_series, pattern):
    # find patterns long-enough
    beam_mode_spfr_ramp_pattern_df, _ = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        beam_mode_with_pattern_series, pattern=list(pattern.values()))
    beam_mode_spfr_ramp_pattern_df['is_sb_long_enough'] = (beam_mode_spfr_ramp_pattern_df[14.0] -
                                                           beam_mode_spfr_ramp_pattern_df[8.0]) > 1800 * 1e9
    beam_mode_spfr_ramp_pattern_long_enough_df = beam_mode_spfr_ramp_pattern_df.query('is_sb_long_enough == True')

    t_setups = beam_mode_spfr_ramp_pattern_long_enough_df[2].values
    t_preramps = beam_mode_spfr_ramp_pattern_long_enough_df[6].values
    t_flattops = beam_mode_spfr_ramp_pattern_long_enough_df[8].values
    t_start_injs = np.array([max(t_s.item(), t_p.item() - int(1800 * 1e9)) for t_s, t_p in zip(t_setups, t_preramps)])
    t_end_injs = np.array([t_preramp.item() for t_preramp in t_preramps])
    t_start_sbs = np.array([t_flattop.item() for t_flattop in t_flattops])
    t_end_sbs = np.array([t_flattop.item() + int(1800 * 1e9) for t_flattop in t_flattops])
    busbar_df = pd.DataFrame({'t_start_injs': t_start_injs,
                              't_end_injs': t_end_injs,
                              't_start_sbs': t_start_sbs,
                              't_end_sbs': t_end_sbs})
    busbar_df.sort_values(by='t_start_injs', inplace=True)

    return busbar_df


def get_complete_busbar_df(t_start, t_end, pattern, spark):
    # {'SETUP': 2.0, 'PRERAMP': 6.0, 'FLATTOP': 8.0, 'RAMPDOWN': 14.0}

    # query beam mode
    beam_mode_with_pattern_df = QueryBuilder().with_nxcals(spark) \
        .with_duration(t_start=t_start, t_end=t_end) \
        .with_query_parameters(nxcals_system='CMW', signal='HX:BMODE') \
        .signal_query() \
        .filter_values(list(pattern.keys())) \
        .map_values(pattern) \
        .dfs[0]

    # generate busbar status
    busbar_df = generate_busbar_status(beam_mode_with_pattern_df["HX:BMODE"], pattern)

    return busbar_df


def get_busbar_resistances(circuit_type, signal_type, t_start_inj, t_end_inj, t_start_sb, t_end_sb, spark):
    # based on plateau, define udf
    def plateau_translate(timestamp):
        if (timestamp >= t_end_inj) & (timestamp < t_start_sb):
            return 0
        elif (timestamp >= t_start_inj) & (timestamp <= t_end_inj):
            return 1
        elif (timestamp >= t_start_sb) & (timestamp <= t_end_sb):
            return 2
        else:
            return -1

    plateau_translate_udf = F.udf(plateau_translate, IntegerType())

    # query current
    i_meas_feature_df = QueryBuilder() \
        .with_nxcals(spark) \
        .with_duration(t_start=int(t_start_inj), t_end=int(t_end_sb)) \
        .with_circuit_type(circuit_type) \
        .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
        .feature_query(['mean', 'std'], function=plateau_translate_udf) \
        .sort_values(by='class') \
        .df

    # query voltage
    u_res_feature_df = QueryBuilder() \
        .with_nxcals(spark) \
        .with_duration(t_start=int(t_start_inj), t_end=int(t_end_sb)) \
        .with_circuit_type(circuit_type) \
        .with_metadata(circuit_name='*', system='BUSBAR', signal=signal_type, wildcard={'BUSBAR': '*'}) \
        .feature_query(['mean', 'std'], function=plateau_translate_udf) \
        .correct_voltage_sign() \
        .sort_busbar_location(circuit_type, circuit_name='*') \
        .df

    # calculate resistance
    res_busbar_df = ResistanceBuilder().with_busbar_voltage(u_res_feature_df) \
        .with_busbar_current(i_meas_feature_df) \
        .calculate_mean_resistance(circuit_type) \
        .convert_to_row(index=t_start_inj)

    return res_busbar_df
