from ipywidgets import Dropdown, Text, HBox, VBox, Select, Label, ToggleButtons, Layout
import getpass

from lhcsmapi.Time import Time
from lhcsmapi.gui.ModuleMediator import ModuleMediator
from lhcsmapi.gui.pc.FgcBrowserAction import FgcBrowserAction
from lhcsmapi.gui.pc.FgcPmSearchBaseModule import FgcPmSearchBaseModule
from lhcsmapi.gui.qh.DateTimeBaseModule import DateTimeBaseModule
from lhcsmapi.gui.qh.QhBrowserAction import QhBrowserAction
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


class FgcPmSearchModuleMediator(ModuleMediator):

    def __init__(self, date_time_module: DateTimeBaseModule, fgc_pm_search_module: FgcPmSearchBaseModule,
                 circuit_type='RB'):
        self._circuit_type = circuit_type
        # First row
        if circuit_type == 'RB':
            circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        elif circuit_type == 'RQ':
            rq_names = SignalMetadata.get_circuit_names('RQ')
            circuit_names = list(
                {rq_name.replace('D', '') if 'RQD' in rq_name else rq_name.replace('F', '') for rq_name in rq_names})
            circuit_names.sort()
        else:
            KeyError('Circuit type {} is not supported'.format(circuit_type))

        self.select_circuit_name_dropdown = Dropdown(
            options=circuit_names,
            description='Circuit name:',
            disabled=False,
        )

        self._date_time_module = date_time_module
        self._date_time_module.mediator = self

        self.done_by_txt = Text(placeholder='Your name', description='Done by:', value=getpass.getuser(),
                                disabled=True)

        self.analysis_mode_tgl_btn = ToggleButtons(
            options=['Automatic', 'Manual'],
            description='Analysis:',
            disabled=False,
            button_style='',
            tooltips=['Select automatic for automatic execution, comment, and acceptance.',
                      'Select manual to execute, comment, and accept each analysis manually.']
        )

        first_row = HBox(children=[self.select_circuit_name_dropdown, self._date_time_module.widget])

        # Second row
        second_row = HBox(children=[self.analysis_mode_tgl_btn, self.done_by_txt])

        # Third row
        self._fgc_pm_search_module = fgc_pm_search_module
        self._fgc_pm_search_module.mediator = self
        third_row = fgc_pm_search_module.widget

        # Fourth row
        self.fgc_pm_sel = Select(
            options=[],
            rows=10,
            disabled=False,
            layout=Layout(width='50%')
        )
        fourth_row = HBox(children=[Label(value='FGC PM Entries:'), self.fgc_pm_sel])

        # Final GUI
        self.widget = VBox(children=[first_row, second_row, third_row, fourth_row])

    def notify(self, sender: object, event: str) -> None:
        start_date_time = self._date_time_module.get_start_date_time()
        end_date_time = self._date_time_module.get_end_date_time()
        circuit_type = self.get_circuit_type()
        circuit_name = self.get_circuit_name()

        if event == QhBrowserAction.DATE_CHANGED:
            pass
        elif event == FgcBrowserAction.FIND_BUTTON_CLICKED:
            self.fgc_pm_entries = self._fgc_pm_search_module.find_fgc_pm_entries_with_qps(start_date_time,
                                                                                          end_date_time,
                                                                                          circuit_type, circuit_name)
            if circuit_type == 'RB':
                self.fgc_pm_sel.options = [Time.to_string(el[1]) for el in self.fgc_pm_entries]
            elif circuit_type == 'RQ':
                self.fgc_pm_sel.options = [
                    '{}: {}; {}: {}'.format(x[0], Time.to_string(x[1]), x[2], Time.to_string(x[3])) for x in
                    self.fgc_pm_entries]
        else:
            raise KeyError(f'Action {event} not supported!')

    def get_circuit_type(self):
        return self._circuit_type

    def get_circuit_name(self):
        return self.select_circuit_name_dropdown.value

    def is_automatic_mode(self):
        return self.analysis_mode_tgl_btn.value == 'Automatic'

    def get_author(self):
        return self.done_by_txt.value

    def get_fgc_timestamp(self):
        if len(self.fgc_pm_sel.options) == 0:
            raise ValueError(
                'The list of FGC PM Entries should contain at least one element. Please adjust circuit signal as well '
                'as start and end date.')
        else:
            if self.get_circuit_type() == 'RB':
                # ToDo: take the value form the self.fgc_pm_entries to avoid conversion issues
                return Time.to_unix_timestamp(self.fgc_pm_sel.value)
            elif self.get_circuit_type() == 'RQ':
                idx = self.fgc_pm_sel.index
                return self.fgc_pm_entries[idx][1], self.fgc_pm_entries[idx][3]

    def get_fgc_circuit(self):
        if self.get_circuit_type() == 'RB':
            return self.select_circuit_name_dropdown.value
        elif self.get_circuit_type() == 'RQ':
            idx = self.fgc_pm_sel.index
            return self.fgc_pm_entries[idx][0], self.fgc_pm_entries[idx][2]
